;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "vterm-toggle" "20260601.1632"
  "Toggles between the vterm buffer and other buffers."
  '((emacs "25.1")
    (vterm "0.0.1"))
  :url "https://github.com/jixiuf/vterm-toggle"
  :commit "a0051a8b8eaa85f8df54ddb032f5710c1f62779d"
  :revdesc "a0051a8b8eaa"
  :keywords '("vterm" "terminals")
  :authors '((nil . "jixiufjixiuf@qq.com"))
  :maintainers '((nil . "jixiufjixiuf@qq.com")))
